#include "pgm.h"


/**
 * pgm_read_from_file reads a pgm file and fills the pgm struct "p" with its data
 * @param p is a pointer of pgm struct
 * @param filename is a pointer of characters representing the path of the targeted file
 * @return pgm_error
**/
pgm_error pgm_read_from_file(pgm *p, char *filename){
    if (filename == NULL || p == NULL){
        return failure;
    }

    char header[100];
    int px_lin, px_col, px_max;
    FILE *fp;
    fp = fopen(filename, "rb");
    fscanf(fp, "%s %d %d %d", header, &px_lin, &px_col, &px_max);
    matrix_alloc(&(p->pixels), px_col, px_lin);
    p->max = px_max;
    for (int i = 0; i < px_lin; i++){
        for (int j = 0; j < px_col; j++){
            uint8_t tmp;
            fread(&tmp, sizeof(uint8_t), 1, fp);
            p->pixels.data[i][j] = (int32_t)tmp;
        }
    }
    fclose(fp);
    return success;
}


/**
 * pgm_write_to_file creates a pgm file from a pgm struct with a custom name
 * @param p is a pointer of pgm struct
 * @param filename is a pointer of characters that defines the custom file name
 * @return pgm_error
**/
pgm_error pgm_write_to_file(pgm *p, char *filename){
    if (filename == NULL || p == NULL){
        return failure;
    }

    FILE *nfp;
    nfp = fopen(filename, "wb");

    fprintf(nfp, "P5\n");
    fprintf(nfp, "%d %d\n", p->pixels.c, p->pixels.l);
    fprintf(nfp, "%d\n", p->max);


    for (int i = 0; i < p->pixels.c; i++){
        for (int j = 0; j < p->pixels.l; j++){
            fputc(p->pixels.data[i][j], nfp);
        }
    }
    fclose(nfp);
    return success;
}


/**
 * pgm_negative modifies an empty pgm struct with a negative version of the original one
 * @param neg is a pointer of pgm struct which contains the negative version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error pgm_negative(pgm *neg, const pgm *const orig){
    if (neg == NULL || orig == NULL){
        return failure;
    }

    neg->max = orig->max;
    matrix_alloc(&neg->pixels, orig->pixels.c, orig->pixels.l);

    for (int i = 0; i < orig->pixels.c; i++){
        for (int j = 0; j < orig->pixels.l; j++){
            neg->pixels.data[i][j] = 255 - orig->pixels.data[i][j];
        }
    }
    return success;
}


/**
 * pgm_symmetry_hori modifies an empty pgm struct with a horizontal symmetric version of the original one
 * @param sym is a pointer of pgm struct which contains the modified version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error pgm_symmetry_hori(pgm *sym, const pgm *const orig){
    if (sym == NULL || orig == NULL){
        return failure;
    }

    sym->max = orig->max;
    matrix_alloc(&sym->pixels, orig->pixels.c, orig->pixels.l);

    for (int i = 0; i < orig->pixels.c; i++){
        for (int j = 0; j < orig->pixels.l; j++){  
            sym->pixels.data[i][j] = orig->pixels.data[i][(orig->pixels.l - 1) - j];
        }
    }
    return success;
}


/**
 * pgm_symmetry_vert modifies an empty pgm struct with a vertical symmetric version of the original one
 * @param sym is a pointer of pgm struct which contains the modified version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error pgm_symmetry_vert(pgm *sym, const pgm *const orig){
     if (sym == NULL || orig == NULL){
        return failure;
    }

    sym->max = orig->max;
    matrix_alloc(&sym->pixels, orig->pixels.c, orig->pixels.l);

    for (int i = 0; i < orig->pixels.c; i++){
        for (int j = 0; j < orig->pixels.l; j++){  
            sym->pixels.data[i][j] = orig->pixels.data[(orig->pixels.l - 1) - i][j];
        }
    }
    return success;
}


/**
 * pgm_symmetry_cent modifies an empty pgm struct with a central symmetric version of the original one
 * @param sym is a pointer of pgm struct which contains the modified version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error pgm_symmetry_cent(pgm *sym, const pgm *const orig){
     if (sym == NULL || orig == NULL){
        return failure;
    }

    sym->max = orig->max;
    matrix_alloc(&sym->pixels, orig->pixels.c, orig->pixels.l);

    for (int i = 0; i < orig->pixels.c; i++){
        for (int j = 0; j < orig->pixels.l; j++){  
            sym->pixels.data[i][j] = orig->pixels.data[(orig->pixels.l - 1) - i][(orig->pixels.l - 1) - j];
        }
    }
    return success;
}


/**
 * pgm_photomaton modifies an empty pgm struct with a x4 photomaton version of the original one
 * @param photomaton is a pointer of pgm struct which contains the modified version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error pgm_photomaton(pgm *photomaton, const pgm *const orig){
    if (photomaton == NULL || orig == NULL){
        return failure;
    }
    int newI = -1;
    int newJ = -1;
    photomaton->max = orig->max;
    matrix_alloc(&photomaton->pixels, orig->pixels.c, orig->pixels.l);

    for (int i = 0; i < orig->pixels.l; i++){
        for (int j = 0; j < orig->pixels.c; j++){  
            if (i % 2 != 0){
                newI = ((i - 1) / 2) + (orig->pixels.l / 2);
            }else{newI = i / 2;}
        
            if (j % 2 != 0){
                newJ = ((j - 1) / 2) + (orig->pixels.c / 2);
            }else{newJ = j / 2;}

            photomaton->pixels.data[newI][newJ] = orig->pixels.data[i][j];
        }
    }
    return success;
}


/**
 * pgm_crop modifies an empty pgm struct with a cropped version of the original one depending on bounds
 * @param crop is a pointer of pgm struct which contains the cropped version
 * @param orig constant pointer of pgm which is the original pgm
 * @param x0 is the lower x axis bound
 * @param x1 is the upper x axis bound
 * @param y0 is the lower y axis bound
 * @param y1 is the upper y axis bound
 * @return pgm_error
 * 
 * WARNING: this function doesn't work properly, only squared bounds make it work
**/
pgm_error pgm_crop(pgm *crop, const pgm *const orig, int32_t x0, int32_t x1, int32_t y0, int32_t y1){
    if (crop == NULL || orig == NULL){
        return failure;
    }
    if (x0 < 0 || y0 < 0 || x0 > x1 || y0 > y1 || x1 > orig->pixels.c || y1 > orig->pixels.l){
        return failure;
    }
    //Still buggy for rectangles 
    matrix_extract_submatrix(&crop->pixels, orig->pixels, x0, x1, y0, y1);
    crop->max = orig->max;
    return success;
}


/**
 * add_zeros_in_pgm modifies an empty pgm struct with a "zero bordered" version of the original one
 * Used to avoid edge problems when calculating convolutions
 * @param zeroedPgm is a pointer of pgm struct which contains the "zero bordered" version
 * @param orig constant pointer of pgm which is the original pgm
 * @return pgm_error
**/
pgm_error add_zeros_in_pgm(pgm *zeroedPgm, const pgm *const orig){
    if (zeroedPgm == NULL || orig == NULL){
        return failure;
    }
    zeroedPgm->max = orig->max;
    matrix_alloc(&zeroedPgm->pixels, (orig->pixels.l + 2), (orig->pixels.c + 2));
    
    for(int i = 0; i < zeroedPgm->pixels.c; i++){
        for (int j = 0; j < zeroedPgm->pixels.l; j++){
            zeroedPgm->pixels.data[i][j] = 0;
        }
    }

    for(int i = 1, k = 0; i < zeroedPgm->pixels.c - 1; i++, k++){
        for(int j = 1, m = 0; j < zeroedPgm->pixels.l - 1; j++, m++){
            zeroedPgm->pixels.data[i][j] = orig->pixels.data[k][m];
        }
    }

    return success;
}


/**
 * calculate_pixel_conv calculates the convoluted version of a pixel using a sub matrix from 
 * an image and a kernel
 * @param m is a matrix containing pixels from an image
 * @param kernel constant pointer of matrix containing values to calcuate a convolution
 * @param max represents the maximum value of a pixel
 * @return result
**/
int calculate_pixel_conv(const matrix m, const matrix *const kernel, int max){
    int result = 0;

    for (int i = 0; i < m.c; i++){
        for (int j = 0; j < m.l; j++){
            result += m.data[i][j] * kernel->data[i][j];
        }
    }
    result /= (kernel->l * kernel->c);

    if(result > max - 1){return max - 1;}
    if(result < 0){return 0;}
    return result;
}


/**
 * pmg_conv modifies an empty pgm struct with a "convoluted" version of the original one depending on a kernel
 * @param conv is a pointer of pgm which contains the "convoluted" version
 * @param orig constant pointer of pgm which is the original pgm
 * @param kernel constant pointer of matrix containing values to calcuate a convolution
 * @param norm double number added to modify convolution "behaviour"
 * @return pgm_error
**/
pgm_error pmg_conv(pgm *conv, const pgm *const orig, const matrix *const kernel, double norm){
    if (conv == NULL || orig == NULL || kernel == NULL){
        return failure;
    }

    conv->max = orig->max;
    matrix_alloc(&conv->pixels, orig->pixels.l - 2, orig->pixels.c - 2);
    
    for (int i = 1; i < orig->pixels.c - 2; i++){
        for (int j = 1; j < orig->pixels.l - 2; j++){
            matrix tmp;
            matrix_extract_submatrix(&tmp, orig->pixels, i - 1, i + 2, j - 1, j + 2);
            conv->pixels.data[i][j] = calculate_pixel_conv(tmp, kernel, orig->max) * norm;
            matrix_destroy(&tmp);
        }
    }
    return success;
}


/**
 * demo uses SDL2 library to display an image on a window
 * @param img is a constant pointer of the pgm displayed on screen 
 * @return pgm_error
**/
pgm_error demo(const pgm *const img){
    int width = img->pixels.c, height = img->pixels.l;
	struct gfx_context_t *ctxt = gfx_create("Example", width, height);

    while (gfx_keypressed() != 27/*ESC btn code*/) {
		for (int i = 0; i < img->pixels.c; i++) {
           for (int j = 0; j < img->pixels.l; j++) {
                gfx_putpixel(ctxt, j, i, MAKE_COLOR(img->pixels.data[i][j], img->pixels.data[i][j], img->pixels.data[i][j]));
           }
	    }
		gfx_present(ctxt);
	}
    gfx_destroy(ctxt);
    return success;
}